# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../nbs/agents/benchmark_agents/99_TEMP_PPO_SB3.ipynb.

# %% auto 0
__all__ = []

# %% ../../../nbs/agents/benchmark_agents/99_TEMP_PPO_SB3.ipynb 3
# General libraries:
import numpy as np
from scipy.stats import norm
from tqdm import tqdm

# TEMP libraries:
import gym
from stable_baselines3 import PPO
from stable_baselines3.common.env_util import make_vec_env

# Mushroom libraries
from mushroom_rl.core import Agent



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../../nbs/agents/rl_agents/pre_specified_agents/32_SAC_gumbel.ipynb.

# %% auto 0
__all__ = ['SACGumbel']

# %% ../../../../nbs/agents/rl_agents/pre_specified_agents/32_SAC_gumbel.ipynb 4
# General libraries
import numpy as np

# Torch
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim

# Networks
from ...networks.critics import CriticNetworkStateAction
from ...networks.actors import ActorNetwork

# Algorithms
from ..sac_gumbel import SAC_gumbel

# Processors
from ...processors.processors import  OneHotToDiscrete


# %% ../../../../nbs/agents/rl_agents/pre_specified_agents/32_SAC_gumbel.ipynb 5
class SACGumbel():

    """
    Soft Actor Critic (SAC) agent for discrete action space. Follows the implementation from https://arxiv.org/pdf/1910.07207.pdf

    Args:
        mdp_info (MDPInfo): Contains relevant information about the environment.
        num_actions (int): Number of actions of the environment. If None, it will be obtained from the environment.
        learning_rate_actor (float): Learning rate for the actor.
        learning_rate_critic (float): Learning rate for the critic.
        learning_rate_alpha (float): Learning rate for the temperature parameter.
        initial_replay_size (int): Number of transitions to save in the replay buffer during startup.
        max_replay_size (int): Maximum number of transitions to save in the replay buffer.
        batch_size (int): Number of transitions to sample each time experience is replayed.
        n_features (int): Number of features for the hidden layers of the networks.
        warmup_transitions (int): Number of transitions to replay before starting to update the networks.
        lr_alpha (float): Learning rate for the temperature parameter.
        tau (float): Parameter for the soft update of the target networks.
        target_entropy (float): Target entropy - follows rule of thumb of 0.98 * -log(1/|A|) if None, otherwise it uses the provided value.
        temperature (float): Temperature of the Gumbel Softmax distribution for backpropagation.
        optimizer (torch.optim): Optimizer to use for the networks.
        squeeze_output (bool): Whether to squeeze the output of the actor network or not.
        use_cuda (bool): Whether to use CUDA or not. If True and not available, it will use CPU.
        agent_name (str): Name of the agent. If set to None will use some default name.

    """

    def __init__(
            self,
            mdp_info,
            num_actions = None,
            learning_rate_actor = 3e-4,
            learning_rate_critic = None,
            learning_rate_alpha = 3e-4,
            initial_replay_size = 64,
            max_replay_size = 50000,
            batch_size = 64,
            n_features = 64,
            warmup_transitions = 100,
            lr_alpha = 3e-4,
            tau = 0.005,
            target_entropy = None,
            temperature = 1.0,
            optimizer = optim.Adam,
            squeeze_output = True,
            use_cuda = True,
            agent_name = None):
        
        use_cuda = use_cuda and torch.cuda.is_available()

        actor_input_shape = mdp_info.observation_space.shape
        
        if num_actions == None:
            actor_output_shape = (mdp_info.action_space.shape[0],)
        else:
            actor_output_shape = (num_actions,)

        critic_input_shape = (actor_input_shape[0] + actor_output_shape[0],)

        if learning_rate_critic is None:
            learning_rate_critic = learning_rate_actor

        if target_entropy is None:
            target_entropy = -np.log((1.0 / mdp_info.action_space.shape[0])) * 0.98

        actor_params = dict(network=ActorNetwork,
                                n_features=n_features,
                                input_shape=actor_input_shape,
                                output_shape=actor_output_shape,
                                use_cuda=use_cuda)

        actor_optimizer = {'class': optimizer,
                    'params': {'lr': learning_rate_alpha}} 
        
        critic_params = dict(network=CriticNetworkStateAction,
                                optimizer={'class': optim.Adam,
                                        'params': {'lr': 3e-4}}, # default: 3e-4
                                loss=F.mse_loss,
                                n_features=n_features,
                                input_shape=critic_input_shape,
                                output_shape=(1,),
                                squeeze_output=squeeze_output,
                                use_cuda=use_cuda)
    
        # Agent
        self.agent = SAC_gumbel(mdp_info, actor_params,
                    actor_optimizer, critic_params, batch_size, initial_replay_size,
                    max_replay_size, warmup_transitions, tau, lr_alpha, temperature = temperature,
                    critic_fit_params=None, target_entropy=target_entropy)
        
        if agent_name is None:
            self.agent.name = 'SAC_gumbel'
        else:
            self.agent.name = agent_name

        self.agent.add_postprocessor(OneHotToDiscrete())

    def __getattr__(self, attr):
        return getattr(self.agent, attr)





# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../nbs/agents/benchmark_agents/04_Gradient_boosting_NV_agents.ipynb.

# %% auto 0
__all__ = ['FakePolicy', 'XGBAgent', 'XGBMetaAgent']

# %% ../../../nbs/agents/benchmark_agents/04_Gradient_boosting_NV_agents.ipynb 4
# General libraries:
import numpy as np
from scipy.stats import norm
from tqdm import trange, tqdm
from time import sleep

import xgboost as xgb

from ..processors.processors import GetTimeSeriesAndStaticFeatures

# Mushroom libraries
from mushroom_rl.core import Agent

from timeit import default_timer as timer

from .nv_agents_ERM import NewsvendorDataMeta
from torch.utils.data import DataLoader

# %% ../../../nbs/agents/benchmark_agents/04_Gradient_boosting_NV_agents.ipynb 6
class FakePolicy():
    def reset():
        pass

class XGBAgent(Agent):
    def __init__(self,

                 ### Pinball loss params
                 cu=None,
                 co=None,

                 ### XGB params
                 eta=0.3,
                 gamma=0,
                 max_depth=6,
                 min_child_weight=1,
                 max_delta_step=0,
                 subsample=1,
                 sampling_method="uniform",
                 colsample_bytree=1,
                 colsample_bylevel=1,
                 colsample_bynode=1,
                 lambda_=1,
                 alpha=0,
                 tree_method="auto",
                 scale_pos_weight=1,
                 # updater will always use default
                 refresh_leaf=1,
                 # process type will always use default
                 grow_policy="depthwise",
                 max_leaves=0,
                 max_bin=256,
                 num_parallel_tree=1,
                 multi_strategy="one_output_per_tree",
                 max_cached_hist_node=65536,
 
                 ### General params
                 nthread=1,
                 device="cpu",
                 agent_name="XGB_quantile",
                 ):

        self.cu = cu
        self.co = co

        self.model = xgb.XGBRegressor(
            objective = "reg:quantileerror",
            quantile_alpha = cu / (cu + co),

            eta=eta,
            gamma=gamma,
            max_depth=max_depth,
            min_child_weight=min_child_weight,
            max_delta_step=max_delta_step,
            subsample=subsample,
            sampling_method=sampling_method,
            colsample_bytree=colsample_bytree,
            colsample_bylevel=colsample_bylevel,
            colsample_bynode=colsample_bynode,
            lambda_=lambda_,
            alpha=alpha,
            tree_method=tree_method,
            scale_pos_weight=scale_pos_weight,
            # updater will always use default
            refresh_leaf=refresh_leaf,
            # process type will always use default
            grow_policy=grow_policy,
            max_leaves=max_leaves,
            max_bin=max_bin,
            num_parallel_tree=num_parallel_tree,
            multi_strategy=multi_strategy,
            max_cached_hist_node=max_cached_hist_node,
            
            nthread=nthread,
            device=device
        )

        self.train_directly=True
        self.train_mode = "direct"
        self.policy = FakePolicy

        self._postprocessors=list()
        self._preprocessors=list() 

        self.name = agent_name
        self.fitted=False
        
    def _get_fitted_model(self, X, y, mask=None):

        self.model.fit(X, y)
        self.n_features_ = X.shape[1]
    

    def fit(self, features, demand, mask=None):

        X=features
        y=demand

        # X = np.random.normal(0, 1, (1, X.shape[1]))
        # y = np.random.normal(0, 1, (y.shape))

        # X = np.random.rand(100, 5)
        # y = np.random.rand(100)

        # print unique values of y

        if y.ndim == 1:
            y = np.reshape(y, (-1, 1))

        self._get_fitted_model(X, y, mask)

        self.fitted=True

        return


    def draw_action(self, X):   

        if self.fitted:  

            if X.ndim == 1:
                X = np.reshape(X, (-1, self.n_features_))

            pred = self.model.predict(X)

        else:
            pred = np.random.rand(1)  
        
        return pred


# %% ../../../nbs/agents/benchmark_agents/04_Gradient_boosting_NV_agents.ipynb 7
class XGBMetaAgent(Agent):
    def __init__(self,
                 feature_map,
                 input_size,
                 lag_features,

                 ### Pinball loss params
                 cu,
                 co,
                 num_time_series_features,
                 lag_window,
                 agent_name="XGBMeta_quantile",

                 ### XGB params
                 eta=0.3,
                 gamma=0,
                 max_depth=6,
                 min_child_weight=1,
                 max_delta_step=0,
                 subsample=1,
                 sampling_method="uniform",
                 colsample_bytree=1,
                 colsample_bylevel=1,
                 colsample_bynode=1,
                 lambda_=1,
                 alpha=0,
                 tree_method="auto",
                 scale_pos_weight=1,
                 # updater will always use default
                 refresh_leaf=1,
                 # process type will always use default
                 grow_policy="depthwise",
                 max_leaves=0,
                 max_bin=256,
                 num_parallel_tree=1,
                 multi_strategy="one_output_per_tree",
                 max_cached_hist_node=65536,
 
                 ### General params
                 nthread=1,
                 device="cpu",
                 ):

        self.cu = cu
        self.co = co

        self.feature_map = feature_map
        self.num_time_series_features = num_time_series_features
        self.lag_window = lag_window
        self.lag_features = lag_features
        self.input_size = input_size  

        self.model = xgb.XGBRegressor(
            objective = "reg:quantileerror",
            quantile_alpha = cu[0] / (cu[0] + co[0]), # Currently only works if cu and co are the same for each product

            eta=eta,
            gamma=gamma,
            max_depth=max_depth,
            min_child_weight=min_child_weight,
            max_delta_step=max_delta_step,
            subsample=subsample,
            sampling_method=sampling_method,
            colsample_bytree=colsample_bytree,
            colsample_bylevel=colsample_bylevel,
            colsample_bynode=colsample_bynode,
            lambda_=lambda_,
            alpha=alpha,
            tree_method=tree_method,
            scale_pos_weight=scale_pos_weight,
            # updater will always use default
            refresh_leaf=refresh_leaf,
            # process type will always use default
            grow_policy=grow_policy,
            max_leaves=max_leaves,
            max_bin=max_bin,
            num_parallel_tree=num_parallel_tree,
            multi_strategy=multi_strategy,
            max_cached_hist_node=max_cached_hist_node,
            
            nthread=nthread,
            device=device
        )

        self.train_directly=True
        self.train_mode = "direct"
        self.policy = FakePolicy

        self._postprocessors=list()
        self._preprocessors=list() 

        self.name = agent_name
        self.fitted=False
        
    def _get_fitted_model(self, X, y, mask=None):

        self.model.fit(X, y)
        self.n_features_ = X.shape[1]
    

    def fit(self, features, demand, mask=None):

        print("separating features")

        features_time_dependent = features[0]
        features_timeless = features[1]

        print("setting up dataset")

        dataset_train = NewsvendorDataMeta(features_time_dependent, features_timeless, demand, self.feature_map, self.lag_window, self.input_size, self.lag_features)


        batch_size = int(len(dataset_train))
        print("batch_size:", batch_size)
        train_loader=DataLoader(dataset=dataset_train, batch_size=batch_size, shuffle=True, num_workers=1,) # make shuffle true

        print("loading data")
        start_time = timer()

        for i, batch in enumerate(train_loader): # will only be one batch
            X = batch[0].numpy()
            X = np.reshape(X, (X.shape[0], -1))
            y = batch[1].numpy()
            product = batch[2]

            if i >=0:
                break

        end_time = timer()

        print("loaded data in seconds:", end_time-start_time)
        
        print(X.shape)
        print(y.shape)

        # print size of X and y in GB:
        print("size X in GB:", X.nbytes / 1e9)
        print("size y in GB:", y.nbytes / 1e9)

        if y.ndim == 1:
            y = np.reshape(y, (-1, 1))

        print(y.shape)

        start_time = timer()
        self._get_fitted_model(X, y, mask)
        end_time = timer()
        print("fitted model in seconds:", end_time-start_time)

        self.fitted=True

        return


    def draw_action(self, X):   

        if self.fitted:  

            if X.ndim == 1:
                X = np.reshape(X, (-1, self.n_features_))
            
            if len(X.shape) == 3:
                X = np.reshape(X, (X.shape[0], -1))

            pred = self.model.predict(X)

        else:
            pred = np.random.rand(1)  
        
        return pred



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/environments/10_calculation_functions.ipynb.

# %% auto 0
__all__ = ['normalize_reward_FC', 'normalize_reward_NV', 'get_orders_arriving_stochastic', 'eoq_qr_calculations_single_action',
           'eoq_qr_calculations_qr_action', 'mpfc_calculations_ss_action', 'nv_calculations', 'get_fixed_ordering_cost']

# %% ../../nbs/environments/10_calculation_functions.ipynb 4
# General libraries:
import numpy as np
import time

# %% ../../nbs/environments/10_calculation_functions.ipynb 8
# helper functions:
def normalize_reward_FC(self, reward):  

    # print(self.normalize_reward)
    # print("normalize_reward")
    # print(self.normalize_reward)
    if self.normalize_reward:
        # print("doning normalization")
        # print(np.sum(self.holding_cost*self.inventory_cap))
        # print(np.sum(self.fixed_ordering_cost))
        reward_norm = reward / np.maximum(np.sum(self.holding_cost*self.inventory_cap), np.sum(self.fixed_ordering_cost))
        reward_norm = np.clip(reward_norm, -10, 0)
        if reward_norm <= self.min_reward:
            self.min_reward = reward_norm
            print("before:", reward, "after:", reward_norm, "with:", (np.maximum(np.sum(self.holding_cost*self.inventory_cap), np.sum(self.fixed_ordering_cost))))
    else:
        # print("not normalizing")
        reward_norm = reward

    return reward_norm

# %% ../../nbs/environments/10_calculation_functions.ipynb 9
# helper functions:
def normalize_reward_NV(self, reward):

    if self.normalize_reward:

        reward_norm = reward / np.maximum(np.sum(self.underage_cost*self.avg_demand), np.sum(self.overage_cost*self.avg_demand))

        reward_norm = np.clip(reward_norm, -10, 0)

        if reward_norm <= self.min_reward:
            self.min_reward = reward_norm
            print("before:", reward, "after:", reward_norm, "with:", (np.maximum(np.sum(self.underage_cost*self.avg_demand), np.sum(self.overage_cost*self.avg_demand))))
    else:
        # print("not normalizing")
        reward_norm = reward

    return reward_norm

# %% ../../nbs/environments/10_calculation_functions.ipynb 10
def get_orders_arriving_stochastic(self, action):


    # print("xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx")
    # print("initial state")

    # print("lead times:", self.lead_time_realizations)
    # print("full order pipeline:", self.full_order_pipeline)
    # print("order pipeline:", self.order_pipeline)

    # print("order placed:", action)
    
    all_orders_arriving = np.zeros(self.num_products) 
            
    for i in range(self.num_products):
        arriving_mask = self.lead_time_realizations[i] == 0
        orders_arriving = np.sum(self.full_order_pipeline[i][arriving_mask])
        self.full_order_pipeline[i][arriving_mask] = 0
        all_orders_arriving[i] = orders_arriving
    
    # print("arriving orders:", all_orders_arriving)
    
    self.full_order_pipeline[:,:-1] = self.full_order_pipeline[:,1:].copy()
    self.full_order_pipeline[:,-1] = action

    # print("full order pipeline after update1:", self.full_order_pipeline)

    self.lead_time_realizations[:,:-1] = self.lead_time_realizations[:,1:].copy()

    if self.lead_time_stochastic == "gamma":
        lead_time_drawn_raw = np.random.gamma(self.lead_time-1, 1, self.num_products)+1
    elif self.lead_time_stochastic == "normal":
        lead_time_drawn_raw = np.random.normal(self.lead_time, self.lead_time_variance, self.num_products)
        lead_time_drawn_raw = np.maximum(lead_time_drawn_raw, 1)
    else:
        raise ValueError("lead time distribution unknown")
    
    self.lead_time_realizations[:,-1] = int(np.round(np.minimum(lead_time_drawn_raw, self.max_lead_time), 0))
    # print("lead time of order:", self.lead_time_realizations[:,-1])
    self.lead_time_realizations = np.maximum(self.lead_time_realizations-1, 0)

    # below also accounts for the c  ase that lag window is larger than max lead time

    # print("full order pipeline after update2:", self.full_order_pipeline)

    if self.max_lead_time > self.max_observation_lead_time:
        self.order_pipeline = self.full_order_pipeline[:,-self.max_observation_lead_time:].copy()
    else:
        self.order_pipeline[:,-self.max_lead_time:] = self.full_order_pipeline.copy()


    # print("full order pipeline after update3:", self.full_order_pipeline)

    self.order_pipeline[:,0] = self.order_pipeline[:,0] + np.sum(self.full_order_pipeline[:,:-self.max_observation_lead_time], axis = 1) 


    # print("final state")

    # print("lead times:", self.lead_time_realizations)  
    # print("full order pipeline:", self.full_order_pipeline)
    # print("order pipeline:", self.order_pipeline)


    # time.sleep(10)

    return all_orders_arriving

# %% ../../nbs/environments/10_calculation_functions.ipynb 11
def eoq_qr_calculations_single_action(self, action):

    # TODO Think about naming of the function

    # print(f"action: {action}")
    # print(f"self.inventory: {self.inventory}")
    # # print pipeline
    # print(f"self.order_pipeline: {self.order_pipeline}")
    # # print demand
    # print(f"self.demand: {self.demand}")

    fixed_ordering_cost_step = get_fixed_ordering_cost(self, action, positive_only = True)

    # todo: add variable ordering cost:
        # variable cost positive order ...
        # variable cost negative order would build on overage cost

    # print("################# new step #################")

    # print("start inventory:", self.inventory)
    # print("start pipeline:", self.order_pipeline)

    # print("pipeline last element:" self.order_pipeline[:,-1])

    # print("action:", action)

    order = action

    if self.use_order_pipeline:
        if self.lead_time_stochastic:

            orders_arriving = get_orders_arriving_stochastic(self, order)
         
        else:
            orders_arriving = self.order_pipeline[:,-1].copy()
            # print("orders_arriving: ", orders_arriving)
            #Move elements one slot to the right in order_pipeline
            self.order_pipeline[:, 1:] = self.order_pipeline[:, :-1]
            product_indices = np.arange(self.num_products)
            insert_indices = -self.lead_time
            self.order_pipeline[product_indices, insert_indices] = order

    else:
        orders_arriving = order

    # print("orders_arriving:", orders_arriving)
    # print("new pipeline:", self.order_pipeline)

    self.inventory += orders_arriving # first, the order will be added to inventory (meaning lead-time is 0)

    # print("inventory after order:", self.inventory)

    self.inventory = np.maximum(self.inventory, 0) # inventory cannot be negative

    self.inventory = np.minimum(self.inventory, self.inventory_cap) # inventory cannot be higher than inventory_cap

    # print("demand:", self.demand[self.period, :])

    self.inventory -= self.demand[self.period, :] # Then demand is subtracted from inventory

    outages = np.where(self.inventory < 0, np.abs(self.inventory), 0)
    penalty_cost_step = self.underage_cost * outages.sum()

    self.inventory = np.maximum(self.inventory, 0) # inventory cannot be negative
    holding_cost_step = self.holding_cost * self.inventory

    total_cost_step = fixed_ordering_cost_step + penalty_cost_step + holding_cost_step
    reward = - total_cost_step.sum() # maximize negative cost (sum over all products)

    # print(f"fixed_ordering_cost_step: {fixed_ordering_cost_step}")
    # print(f"penalty_cost_step: {penalty_cost_step}")
    # print(f"holding_cost_step: {holding_cost_step}")


    # print(f"fixed_ordering_cost_step: {fixed_ordering_cost_step}")
    # print(f"penalty_cost_step: {penalty_cost_step}")
    # print(f"holding_cost_step: {holding_cost_step}")
    # print(f"total_cost_step: {total_cost_step}")

    reward_norm = normalize_reward_FC(self, reward)

    # print("new inventory:", self.inventory)

    # # print(f"reward_norm: {reward_norm}")
    # if self.normalize_reward:
    # time.sleep(2)

    info = dict(
            inventory = self.inventory,
            demand = self.demand[self.period, :],
            action = action,
            order = action,
            penalty = penalty_cost_step,
            holding = holding_cost_step,
            fixed_ordering = fixed_ordering_cost_step,
    )

    self.period += 1

    return reward_norm, info

# %% ../../nbs/environments/10_calculation_functions.ipynb 12
def eoq_qr_calculations_qr_action(self, action):

    # TODO Think about naming of the function
    
    # print("inventory: ", self.inventory)
    # print("pipeline: ", self.order_pipeline)
    # print("action: ", action.shape)
    # print("demand: ", self.demand[self.period, :])

    if len(action.shape) == 1:
        r = action[:self.num_products]
        q = action[self.num_products:]

    # print("r: ", r.shape)
    # print("q: ", q.shape)

    order = np.where(self.inventory <= r, q, 0)

    fixed_ordering_cost_step = get_fixed_ordering_cost(self, order, positive_only = True)

    # print(fixed_ordering_cost_step.shape)

    # print("pipeline becfore: ", self.order_pipeline)

    if self.use_order_pipeline:

        if self.lead_time_stochastic:
            orders_arriving = get_orders_arriving_stochastic(self, order)

        else:
            orders_arriving = self.order_pipeline[:,-1].copy()
            # print("orders_arriving: ", orders_arriving)
            #Move elements one slot to the right in order_pipeline
            self.order_pipeline[:, 1:] = self.order_pipeline[:, :-1]
            product_indices = np.arange(self.num_products)
            insert_indices = -self.lead_time
            self.order_pipeline[product_indices, insert_indices] = order
    else:
        orders_arriving = order

    # print("pipeline after: ", self.order_pipeline)

    # print(np.round(self.order_pipeline,0))




    # todo: add variable ordering cost:
        # variable cost positive order ...
        # variable cost negative order would build on overage cost

    inventory_start = self.inventory.copy()

    self.inventory += orders_arriving # first, the order will be added to inventory (meaning lead-time is 0)

    np.set_printoptions(suppress=True)

    self.inventory = np.maximum(self.inventory, 0) # inventory cannot be negative

    self.inventory = np.minimum(self.inventory, self.inventory_cap) # inventory cannot be higher than inventory_cap

    self.inventory -= self.demand[self.period, :] # Then demand is subtracted from inventory

    outages = np.where(self.inventory < 0, np.abs(self.inventory), 0)
    penalty_cost_step = self.underage_cost * outages


    self.inventory = np.maximum(self.inventory, 0) # inventory cannot be negative
    holding_cost_step = self.holding_cost * self.inventory

    total_cost_step = fixed_ordering_cost_step + penalty_cost_step + holding_cost_step
    reward = - total_cost_step.sum() # maximize negative cost (sum over all products)

    # print("end inventory: ", self.inventory)
    # print("end pipeline: ", self.order_pipeline)

    # print(f"fixed_ordering_cost_step: {fixed_ordering_cost_step}")
    # print(f"penalty_cost_step: {penalty_cost_step}")
    # print(f"holding_cost_step: {holding_cost_step}")
    # print(f"total_cost_step: {total_cost_step}")

    reward_norm = normalize_reward_FC(self, reward)


    # print(f"reward_norm: {reward_norm}")

    # if self.normalize_reward:


    info = dict(
            inventory = inventory_start,
            demand = self.demand[self.period, :],
            action = action,
            order = order,
            penalty = penalty_cost_step,
            holding = holding_cost_step,
            fixed_ordering = fixed_ordering_cost_step,
    )

    self.period += 1

    return reward_norm, info

# %% ../../nbs/environments/10_calculation_functions.ipynb 13
def mpfc_calculations_ss_action(self, action):

    # TODO Think about naming of the function

    if len(action.shape) == 1:
        s = action[0]
        S = action[1]
    else:
        s = action[0][0]
        S = action[0][1]

    total_inventory_position= self.inventory + np.sum(self.order_pipeline, axis = 1)
    q = S - total_inventory_position

    # print("inventory:",  self.inventory)
    # print("order pipeline:", self.order_pipeline)
    # print("total_inventory_position:", total_inventory_position)
    # print("S:", S)
    # print("q:", q)

    order = np.where(self.inventory <= s, q, 0)

    # print("order:", order)

    fixed_ordering_cost_step = get_fixed_ordering_cost(self, order, positive_only = True)

    if self.use_order_pipeline:
        orders_arriving = self.order_pipeline[:,-1].copy()
            # Move elements one slot to the right in order_pipeline
        self.order_pipeline[:, 1:] = self.order_pipeline[:, :-1]
        self.order_pipeline[:, 0] = order
    else:
        orders_arriving = order

    # todo: add variable ordering cost:
        # variable cost positive order ...
        # variable cost negative order would build on overage cost

    inventory_start = self.inventory.copy()

    self.inventory += orders_arriving # first, the order will be added to inventory (meaning lead-time is 0)

    self.inventory = np.maximum(self.inventory, 0) # inventory cannot be negative

    self.inventory = np.minimum(self.inventory, self.inventory_cap) # inventory cannot be higher than inventory_cap

    self.inventory -= self.demand[self.period, :] # Then demand is subtracted from inventory

    outages = np.where(self.inventory < 0, np.abs(self.inventory), 0)
    penalty_cost_step = self.underage_cost * outages.sum()

    self.inventory = np.maximum(self.inventory, 0) # inventory cannot be negative
    holding_cost_step = self.holding_cost * self.inventory

    total_cost_step = fixed_ordering_cost_step + penalty_cost_step + holding_cost_step
    reward = - total_cost_step.sum() # maximize negative cost (sum over all products)

    reward_norm = normalize_reward_FC(self, reward)

    info = dict(
            inventory = inventory_start,
            demand = self.demand[self.period, :],
            action = action,
            order = order,
            penalty = penalty_cost_step,
            holding = holding_cost_step,
            fixed_ordering = fixed_ordering_cost_step,
    )

    self.period += 1

    return reward_norm, info

# %% ../../nbs/environments/10_calculation_functions.ipynb 14
def nv_calculations(self, action):
    
    if self.full_demand:
        demand = self.demand[self.period+self.full_additional_context, :]
    else:
        demand = self.demand[self.period, :]

    # assert that action is not nan

    if self.mask is not None:
        mask = self.mask.copy()
        if len(self.mask.shape)==1:
            mask = mask.reshape(-1,1)
        mask = mask[self.period, :]
        action = action*mask

    # action is minimum of 0 and action
    action = np.maximum(action, 0)

    remaining_demand = demand.copy() - action

    underage_cost = self.underage_cost.reshape(1, -1)  # Reshape to (1, product)
    overage_cost = self.overage_cost.reshape(1, -1) 

    cost = np.where(remaining_demand > 0, underage_cost * remaining_demand, 0)
    cost += np.where(remaining_demand < 0, overage_cost * np.abs(remaining_demand), 0)
    # cost[remaining_demand > 0] = self.underage_cost * remaining_demand[remaining_demand > 0]
    # cost[remaining_demand < 0] = self.overage_cost * np.abs(remaining_demand[remaining_demand < 0])

    reward = -cost.sum()

    # print("demand:", demand, "action:", action, "reward: ", reward)

    reward_norm = normalize_reward_NV(self, reward)

    # print(demand[32], action[32])
    
    info = dict(
        demand = demand,
        cost = cost
    )

    self.period += 1

    return reward_norm, info

# %% ../../nbs/environments/10_calculation_functions.ipynb 16
def get_fixed_ordering_cost(self, action, positive_only = True):
    """ 
    Returns the fixed ordering cost for the period.

    Parameters
    ----------
    action : array
        The action taken in the period.
    positive_only : bool
        If True, only positive orders will be charged the fixed ordering cost. Otherwise, all non-zero orders will be charged the fixed ordering cost.

    Returns
    -------
    fixed_ordering_cost : array
        The fixed ordering cost for the period per product.
    
    """

    if positive_only:
        positive_order = np.where(action > 0, 1, 0)
    else:
        # nonzero = 1, zero = 0
        positive_order = np.where(action != 0, 1, 0)
    return self.fixed_ordering_cost * positive_order
